@@ -171,25 +171,28 @@
         (return-from setf `(progn ,@(sb-c::explode-setq form 'error))))
       (when (atom (setq place (macroexpand-for-setf place env)))
         (return-from setf `(setq ,place ,value-form)))
+      (wrap-if
+       (sb-c::compiling-p env)
+       `(sb-c::with-source-form ,place)
+       (block nil
+         (let ((fun (car place)))
+           (when (and (symbolp fun)
+                      ;; Local definition of FUN precludes global knowledge.
+                      (not (sb-c::fun-locally-defined-p fun env)))
+             (let ((inverse (info :setf :expander fun)))
+               ;; NIL is not a valid setf inverse name, for two reasons:
+               ;;  1. you can't define a function named NIL,
+               ;;  2. (DEFSETF THING () ...) is the long form DEFSETF syntax.
+               (when (typep inverse '(cons symbol))
+                 (return `(,(car inverse) ,@(cdr place) ,value-form))))
+             (awhen (transformable-struct-setf-p place env)
+               (return
+                 (slot-access-transform :setf (list (cadr place) value-form) it)))))
 
-      (let ((fun (car place)))
-        (when (and (symbolp fun)
-                   ;; Local definition of FUN precludes global knowledge.
-                   (not (sb-c::fun-locally-defined-p fun env)))
-          (let ((inverse (info :setf :expander fun)))
-            ;; NIL is not a valid setf inverse name, for two reasons:
-            ;;  1. you can't define a function named NIL,
-            ;;  2. (DEFSETF THING () ...) is the long form DEFSETF syntax.
-            (when (typep inverse '(cons symbol))
-              (return-from setf `(,(car inverse) ,@(cdr place) ,value-form))))
-          (awhen (transformable-struct-setf-p place env)
-            (return-from setf
-              (slot-access-transform :setf (list (cadr place) value-form) it)))))
-
-      (multiple-value-bind (temps vals newval setter)
-          (get-setf-expansion place env)
-        (car (gen-let* (mapcar #'list temps vals)
-                       (gen-mv-bind newval value-form (forms-list setter)))))))
+         (multiple-value-bind (temps vals newval setter)
+             (get-setf-expansion place env)
+           (car (gen-let* (mapcar #'list temps vals)
+                          (gen-mv-bind newval value-form (forms-list setter)))))))))
 
   ;; various SETF-related macros
 
