@@ -104,17 +104,19 @@
 ;;; type weakenings, then look for any predicate that is cheaper.
 (defun maybe-weaken-check (type policy)
   (declare (type ctype type))
-  (typecase type
-    ;; Can't do much functional type checking at run-time
-    (fun-designator-type
-     (specifier-type 'function-designator))
-    (fun-type
-     (specifier-type 'function))
-    (t
-     (ecase (policy policy type-check)
-       (0 *wild-type*)
-       (2 (weaken-values-type type))
-       (3 type)))))
+  (let ((policy (policy policy type-check)))
+    (if (zerop policy)
+        *wild-type*
+        (typecase type
+          ;; Can't do much functional type checking at run-time
+          (fun-designator-type
+           (specifier-type 'function-designator))
+          (fun-type
+           (specifier-type 'function))
+          (t
+           (ecase policy
+             (2 (weaken-values-type type))
+             (3 type)))))))
 
 (defun lvar-types-to-check (types original-types n-required)
   (loop for type in types
@@ -153,7 +155,10 @@
          (lvar (node-lvar cast))
          (dest (and lvar (lvar-dest lvar)))
          mv-vars
-         (n-required (length (values-type-required dtype))))
+         (n-required (if (values-type-p dtype)
+                         (length (values-type-required dtype))
+                         (return-from cast-check-types
+                           (values :simple (aver (eq dtype *empty-type*)))))))
     (aver (not (eq ctype *wild-type*)))
     (cond ((and (null (values-type-optional dtype))
                 (not (values-type-rest dtype)))
@@ -248,8 +253,6 @@
                                                                  hairy-data-vector-ref string-hairy-data-vector-ref
                                                                  hairy-data-vector-set string-hairy-data-vector-set)
                                                                fun-name)))))))
-                        #+(or arm64 x86-64)
-                        ((lvar-fun-is (basic-combination-fun dest) '(values-list)))
                         ;; Not great
                         ((lvar-fun-is (basic-combination-fun dest) '(%%primitive))
                          (destructuring-bind (vop &rest args) (basic-combination-args dest)
@@ -269,19 +272,27 @@
                         (t
                          (values-subtypep (lvar-externally-checkable-type lvar)
                                           (cast-type-to-check cast))))))
+            ;; Turn (the fixnum (the integer x)) into (the fixnum x)
             ((and (cast-p dest)
                   (cast-type-check dest)
                   (atom (lvar-uses (node-lvar cast)))
                   (atom (lvar-uses (cast-value dest)))
-                  (lvar-single-value-p (node-lvar cast))
-                  (cond ((and (values-type-p (cast-asserted-type dest))
-                              (values-type-p (cast-asserted-type cast)))
-                         (values-subtypep (cast-asserted-type dest)
-                                          (cast-asserted-type cast)))
-                        ((not (or (values-type-p (cast-asserted-type dest))
-                                  (values-type-p (cast-asserted-type cast))))
-                         (csubtypep (cast-asserted-type dest)
-                                    (cast-asserted-type cast)))))
+                  (let ((asserted-type (cast-asserted-type cast))
+                        (dest-asserted-type (cast-asserted-type dest)))
+                    (cond ((and (values-type-p asserted-type)
+                                (values-type-p dest-asserted-type)
+                                (= (length (values-type-required asserted-type))
+                                   (length (values-type-required dest-asserted-type)))
+                                (= (length (values-type-optional asserted-type))
+                                   (length (values-type-optional dest-asserted-type)))
+                                (eql (values-type-rest asserted-type)
+                                     (values-type-rest dest-asserted-type)))
+                           (values-subtypep dest-asserted-type
+                                            asserted-type))
+                          ((not (or (values-type-p asserted-type)
+                                    (values-type-p dest-asserted-type)))
+                           (csubtypep dest-asserted-type
+                                      asserted-type)))))
              (setf (cast-asserted-type cast) (cast-asserted-type dest)
                    (cast-type-to-check cast) (cast-type-to-check dest)
                    (cast-%type-check dest) nil)
