@@ -630,8 +630,10 @@
                (block nil
                  (map-equality-constraints (make-vector-length-constraint array-var) index gen
                                            (lambda (op not-p)
-                                             (when (and (eq op '>)
-                                                        (not not-p))
+                                             (when (or (and (eq op '>)
+                                                            (not not-p))
+                                                       (and (eq op '<=)
+                                                            not-p))
                                                (return t))))))
       (reoptimize-node node)
       (setf (combination-info node) 'array-in-bounds-p)))
@@ -733,26 +735,30 @@
                                       ((vector-constraint-eq-p (constraint-y con) x)
                                        (and (vector-constraint-eq-p (constraint-x con) y)
                                             (normalize-not (invert-operator (equality-constraint-operator con))
-                                                           (equality-constraint-not-p con))))))
-                              (find-constraint (var1 var2 ref)
-                                (block nil
-                                  (let ((constraints (block-out (node-block ref))))
-                                    (when constraints
-                                      (do-conset-constraints-intersection (con (constraints
-                                                                                (lambda-var-equality-constraints var1)))
-                                        (multiple-value-bind (op not)
-                                            (relations var1 var2 con)
-                                          (when op
-                                            (return (values op not))))))))))
-                       (let ((op1 (find-constraint var1 var2 ref1))
-                             (op2 (find-constraint var2 var1 ref2)))
-                         (case op1
-                           ((< <=) (and (memq op2 '(<= <))
-                                        (push (list '<= var1) r)
-                                        (push (list '<= var2) r)))
-                           ((> >=) (and (memq op2 '(> >=))
-                                        (push (list '>= var1) r)
-                                        (push (list '>= var2) r))))))))
+                                                           (equality-constraint-not-p con)))))))
+                       (let ((constraints1 (block-out (node-block ref1)))
+                             (constraints2 (block-out (node-block ref2))))
+                         (when (and constraints1 constraints2)
+                           (block nil
+                             (do-conset-constraints-intersection (con (constraints1
+                                                                       (lambda-var-equality-constraints var1)))
+                               (let ((op1 (relations var1 var2 con)))
+                                 (when op1
+                                   (do-conset-constraints-intersection (con (constraints2
+                                                                             (lambda-var-equality-constraints var2)))
+                                     (let ((op2 (relations var2 var1 con)))
+                                       (when op2
+                                         (case op1
+                                           ((< <=)
+                                            (when (memq op2 '(<= <))
+                                              (push (list '<= var1) r)
+                                              (push (list '<= var2) r)
+                                              (return)))
+                                           ((> >=)
+                                            (when (memq op2 '(> >=))
+                                              (push (list '>= var1) r)
+                                              (push (list '>= var2) r)
+                                              (return))))))))))))))))
 
                  r)))))))
 
@@ -916,7 +922,7 @@
                   h))
         (values 0 nil))))
 
-(defoptimizer (vector-subseq* constraint-propagate-result) ((sequence start end) node gen)
+(defoptimizer (vector-subseq constraint-propagate-result) ((sequence start end) node gen)
   (let (c
         (null-p (types-equal-or-intersect (lvar-type end) (specifier-type 'null))))
     (when (eql (lvar-type start) (specifier-type '(eql 0)))
