@@ -944,10 +944,10 @@ also processed as top level forms."
 ;;; then the body is converted as usual.
 ;;;
 ;;; When one of these FUNS is declared dynamic extent, we make a
-;;; cleanup with the ENCLOSE as the MESS-UP node and introduce it into
-;;; the lexical environment to convert the body in. We force NEXT to
-;;; start a block outside of this cleanup, causing cleanup code to be
-;;; emitted when the scope is exited.
+;;; cleanup with a dynamic extent node as the mess-up and introduce it
+;;; into the lexical environment to convert the body in. We force NEXT
+;;; to start a block outside of this cleanup, causing cleanup code to
+;;; be emitted when the scope is exited.
 (defun ir1-convert-fbindings (start next result funs body)
   (let ((enclose-ctran (make-ctran)))
     (enclose start enclose-ctran funs)
@@ -1046,11 +1046,7 @@ other."
                      type)
                     ((compiler-values-specifier-type type))
                     (t
-                     (ir1-convert start next result
-                                  `(progn
-                                     ,value
-                                     (error "Bad type specifier: ~a"
-                                            ',type)))
+                     (ir1-convert start next result value)
                      (return-from the-in-policy)))))
     (cond ((or (eq type *wild-type*)
                (eq type *universal-type*)
@@ -1114,7 +1110,7 @@ care."
   (the-in-policy value-type form **zero-typecheck-policy** start next result))
 
 ;;; THE with some options for the CAST
-(def-ir1-translator the* (((value-type &key context silent-conflict
+(def-ir1-translator the* (((type &key context silent-conflict
                                        derive-type-only
                                        truly
                                        source-form
@@ -1122,9 +1118,16 @@ care."
                                        restart)
                            form)
                           start next result)
-  (let ((value-type (if (ctype-p value-type)
-                        value-type
-                        (values-specifier-type value-type)))
+  (let ((type (cond ((ctype-p type)
+                     type)
+                    ((compiler-values-specifier-type type))
+                    (t
+                     (ir1-convert start next result
+                                  `(progn
+                                     ,form
+                                     (error "Bad type specifier: ~a"
+                                            ',type)))
+                     (return-from ir1-convert-the*))))
         (*current-path* (if source-form
                             (ensure-source-path source-form)
                             (ensure-source-path form)))
@@ -1149,19 +1152,19 @@ care."
                                           (the (not exit) (car new-uses)))
                                          (t
                                           new-uses))
-                                   value-type)))))
+                                   type)))))
           (use-annotations
            (ir1-convert start next result form)
            (when result
              (add-annotation result
-                             (make-lvar-type-annotation :type value-type
+                             (make-lvar-type-annotation :type type
                                                         :source-path *current-path*
                                                         :context context))))
           (t
            (let* ((policy (lexenv-policy *lexenv*))
-                  (cast (the-in-policy value-type form (if truly
-                                                           **zero-typecheck-policy**
-                                                           policy)
+                  (cast (the-in-policy type form (if truly
+                                                     **zero-typecheck-policy**
+                                                     policy)
                                        start next result)))
              (when cast
                (setf (cast-context cast) context)
