@@ -60,24 +60,7 @@
           ;; ftypes are for calls and not definitions.
           (case (and arg-info
                      (arg-info-kind arg-info))
-            (:optional (return))
-            (:keyword
-             ;; ;; KLUDGE: keywords are converted before
-             ;; ;; leaf-defined-type is computed:
-             (map-all-uses (lambda (use)
-                             (when (ref-p use)
-                               (let ((leaf (ref-leaf use)))
-                                 (when (lambda-var-p leaf)
-                                   (let ((home (lambda-home (lambda-var-home leaf))))
-                                     (when (and
-                                            (and home
-                                                 (lambda-optional-dispatch home))
-                                            (memq var
-                                                  (optional-dispatch-arglist (lambda-optional-dispatch home))))
-                                       (loop for set in (lambda-var-sets leaf)
-                                             do
-                                             (assert-lvar-type (set-value set) type policy name))))))))
-                           arg)
+            ((:optional :keyword)
              (return))))
         (assert-lvar-type arg
                           type
@@ -252,22 +235,29 @@
                              collect (cons ep n)
                              and do (setf previous-unused (eq ep main))
                              else if (and last more)
-                             collect (cons main (1+ n))))))
-       (flet ((make-args (n)
+                             collect (cons main (1+ n)))))
+            (optional-checked 0))
+       (flet ((check-types (n)
                 (loop for i below n
                       for temp in temps
                       for var in vars
                       for info = (lambda-var-arg-info var)
+                      for type = (and (>= i optional-checked)
+                                      (cond ((and info
+                                                  (eq (arg-info-kind info) :optional)
+                                                  (neq (leaf-defined-type var) *universal-type*))
+                                             ;; Don't delegate to propagate-to-args or
+                                             ;; it will check default values too, but
+                                             ;; FTYPEs are for calls and not definitions.
+                                             (leaf-defined-type var))
+                                            ((neq (leaf-type var) *universal-type*)
+                                             (leaf-type var))))
+                      when type
                       collect
-                      (if (and info (eq (arg-info-kind info) :optional)
-                               (neq (leaf-defined-type var) *universal-type*))
-                          ;; Don't delegate to to propagate-to-args or
-                          ;; it will check default values too, but
-                          ;; FTYPEs are for calls and not definitions.
-                          `(the* (,(leaf-defined-type var)
-                                  :context ,(lambda-var-%source-name var))
-                                 ,temp)
-                          temp))))
+                      `(the* (,type
+                              :context ,(lambda-var-%source-name var))
+                             ,temp)
+                      finally (setf optional-checked n))))
          `(lambda (,n-supplied ,@temps)
             (declare (type index ,n-supplied)
                      (ignorable ,n-supplied))
@@ -275,11 +265,14 @@
               ,@(loop for ((ep . n) . next) on used-eps
                       collect
                       (cond (next
-                             `((eq ,n-supplied ,n)
-                               (%funcall ,ep ,@(make-args n))))
+                             `((progn
+                                 ,@(check-types n)
+                                 (eq ,n-supplied ,n))
+                               (%funcall ,ep ,@(subseq temps 0 n))))
                             (more
                              (with-unique-names (n-context n-count)
                                `(t
+                                 ,@(check-types n)
                                  ,(if (= max n)
                                       `(multiple-value-bind (,n-context ,n-count)
                                            (%more-arg-context ,n-supplied ,max)
@@ -303,11 +296,12 @@
                                                                 (pop vars)))))))))
                             (t
                              `(t
-                               ;; Arg-checking is performed before this step,
+                               ,@(check-types n)
+                               ;; Arg-count checking is performed before this step,
                                ;; arranged by INIT-XEP-ENVIRONMENT,
                                ;; perform the last action unconditionally,
                                ;; and without this the function derived type will be bad.
-                               (%funcall ,ep ,@(make-args n)))))))))))))
+                               (%funcall ,ep ,@(subseq temps 0 n)))))))))))))
 
 (defun can-ignore-optional-ep (n vars keyp)
   (let ((var (loop with i = n
