@@ -137,18 +137,19 @@
 
 (defmacro define-modular-fun-optimizer
     (name ((&rest lambda-list) kind signedp &key (width (gensym "WIDTH"))
-                                                 result-width)
+                                                 result-width
+                                                 (node (gensym "NODE")))
      &body body)
   (%check-modular-fun-macro-arguments name kind lambda-list)
-  (with-unique-names (call args result-width-name)
+  (with-unique-names (args result-width-name)
     `(setf (gethash ',name (modular-class-funs (find-modular-class ',kind ',signedp)))
-           (lambda (,call ,width ,(or result-width
+           (lambda (,node ,width ,(or result-width
                                    result-width-name))
-             (declare (type basic-combination ,call)
+             (declare (type basic-combination ,node)
                       (type (integer 0) ,width)
                       ,@(unless result-width
                           `((ignore ,result-width-name))))
-             (let ((,args (basic-combination-args ,call)))
+             (let ((,args (basic-combination-args ,node)))
                (when (= (length ,args) ,(length lambda-list))
                  (destructuring-bind ,lambda-list ,args
                    (declare (type lvar ,@lambda-list))
@@ -306,7 +307,45 @@
                                                                            name))))
                                  (setf (lvar-%derived-type (node-lvar node)) nil)
                                  (ir1-optimize-combination node))
-                               (values t did-something over-wide)))))))))
+                               (values t did-something over-wide)))))))
+                 (cast
+                  ;; Cut (logand (+ x 1) m), which is (logand (the integer (+ x 1)) m),
+                  ;; and X can only be an integer for that to be true.
+                  (when (eq (cast-type-to-check node)
+                            (specifier-type 'integer))
+                    (let* ((value (cast-value node))
+                           (combination (lvar-uses value)))
+                      (when (and (lvar-matches value :fun-names '(+ -) :arg-count 2)
+                                 (almost-immediately-used-p (node-lvar combination) combination
+                                                            :flushable t))
+                        (destructuring-bind (a b) (combination-args combination)
+                          (when (or (not (types-equal-or-intersect (lvar-type a)
+                                                                   #1=(specifier-type '(or ratio (complex rational)))))
+                                    (not (types-equal-or-intersect (lvar-type b) #1#)))
+                            (labels ((intersect (type)
+                                       (if (values-type-p type)
+                                           (values-type-intersection type
+                                                                     (specifier-type 'integer))
+                                           (type-intersection type
+                                                              (specifier-type 'integer))))
+                                     (change-cast (lvar)
+                                       (let ((cast (lvar-uses lvar)))
+                                         (when (and (cast-p cast)
+                                                    (not (delay-p cast))
+                                                    (almost-immediately-used-p lvar cast :flushable t))
+                                           (setf (cast-asserted-type cast) (intersect (cast-asserted-type cast)))
+                                           (derive-node-type cast
+                                                             (setf (cast-type-to-check cast)
+                                                                   (intersect (cast-type-to-check cast))))))))
+                              (change-cast a)
+                              (change-cast b)
+
+                              (multiple-value-bind (handled any-change wide)
+                                  (cut-lvar value)
+                                (declare (ignore handled))
+                                (when any-change
+                                  (delete-cast node nil))
+                                (values t t wide)))))))))))
              (cut-lvar (lvar &key head
                         &aux did-something must-insert over-wide)
                "Cut all the LVAR's use nodes. If any of them wasn't handled
@@ -359,7 +398,6 @@
              (cond
                ((eq signedp (cdr w)) (<= width (car w)))
                ((eq signedp nil) (< width (car w))))))
-      (declare (dynamic-extent #'inexact-match))
       (let ((tgt (find-if #'inexact-match twidths)))
         (when tgt
           (return-from best-modular-version
@@ -734,3 +772,39 @@
             `(sb-vm::calc-phash val ,n-temps ,steps)
             form)))))
 )
+
+#+(or arm64 x86-64)
+(progn
+  (defknown sb-vm::truncate-mod64 (sb-vm:signed-word sb-vm:signed-word)
+      (values word sb-vm:signed-word)
+      (foldable flushable movable))
+
+  (defoptimizer (sb-vm::truncate-mod64 derive-type) ((n d) node)
+    (let ((res (truncate-derive-type-optimizer node)))
+      (when res
+        (destructuring-bind (q r) (values-type-required res)
+          (make-values-type  (list (%two-arg-derive-type q
+                                                         (specifier-type `(eql ,(ldb (byte sb-vm:n-word-bits 0) -1)))
+                                                         #'logand-derive-type-aux #'logand)
+                                   r))))))
+
+  (deftransform sb-vm::truncate-mod64 ((n d) * * :node node)
+    (let ((truncate-type (truncate-derive-type-optimizer node)))
+      (if (and truncate-type
+               (values-subtypep truncate-type
+                                (values-specifier-type `(values sb-vm:signed-word t &optional))))
+          `(multiple-value-bind (q r) (truncate n d)
+             (values (logand q ,most-positive-word)
+                     r))
+          (give-up-ir1-transform))))
+
+  (define-modular-fun-optimizer truncate
+      ((n d) :untagged nil :width width :node node)
+    (when (and (= width sb-vm:n-word-bits)
+               (not (values-subtypep (node-derived-type node)
+                                     (values-specifier-type `(values sb-vm:signed-word t &optional))))
+               (csubtypep (lvar-type n) (specifier-type 'sb-vm:signed-word))
+               (csubtypep (lvar-type d) (specifier-type 'sb-vm:signed-word)))
+      'sb-vm::truncate-mod64))
+  (setf (gethash 'sb-vm::truncate-mod64 (modular-class-versions (find-modular-class ':untagged 'nil)))
+        `(truncate ,sb-vm:n-word-bits)))
