@@ -7,10 +7,6 @@ int main(int argc, char *argv[], char *envp[])
 #ifdef TRACE_MMAP_SYSCALLS
     extern FILE* mmgr_debug_logfile;
     mmgr_debug_logfile = fopen("mman.log", "w");
-#endif
-#ifdef START_LDB_SERVICE_THREAD
-    extern void init_ldb_service();
-    init_ldb_service();
 #endif
     initialize_lisp(argc, argv, envp);
     lose("unexpected return from initial thread in main()");
