@@ -981,3 +981,93 @@
                       (unless x
                         (or)))
                    :allow-notes nil))
+
+
+(with-test (:name :check-type-bad-type)
+  (assert
+   (nth-value 2
+              (checked-compile `(lambda (x)
+                                  (check-type x (values t)))
+                               :allow-warnings t))))
+
+(with-test (:name :constant-modification-functions-debug)
+  (assert (nth-value 2
+                     (checked-compile
+                      '(lambda ()
+                        (declare (optimize (debug 2)))
+                        (let ((l '((0) 0 0)))
+                          (incf (car (car l)))
+                          l))
+                      :allow-warnings t)))
+  (checked-compile
+   `(lambda ()
+      (flet ((x (x)
+               (when x
+                 (setf (cadr x) 10))))
+        (x nil)
+        (x (list 1 2))))))
+
+(with-test (:name :the-bad-type-no-runtime-error)
+  (multiple-value-bind (fun fail warn)
+      (checked-compile
+       '(lambda (x)
+         (the (nil) x))
+       :allow-warnings t)
+    (assert fail)
+    (assert warn)
+    (assert (= (funcall fun 1) 1))))
+
+(with-test (:name :case-duplicate-t-no-runtime-error)
+  (multiple-value-bind (fun fail warn)
+      (checked-compile
+       '(lambda (x)
+         (case x
+           (t 1)
+           (t 2)))
+       :allow-warnings t)
+    (assert fail)
+    (assert warn)
+    (assert (= (funcall fun 1) 1)))
+  (multiple-value-bind (fun fail warn)
+      (checked-compile
+       '(lambda (x)
+         (case x
+           (otherwise 1)
+           (t 2)))
+       :allow-warnings t)
+    (assert fail)
+    (assert warn)
+    (assert (= (funcall fun 1) 1)))
+  (multiple-value-bind (fun fail warn)
+      (checked-compile
+       '(lambda (x)
+         (typecase x
+           (otherwise 1)
+           (otherwise 2)))
+       :allow-warnings t)
+    (assert fail)
+    (assert warn)
+    (assert (= (funcall fun 1) 1)))
+  (multiple-value-bind (fun fail warn)
+      (checked-compile
+       '(lambda (x)
+         (case x
+           ()
+           (t 1)))
+       :allow-warnings t)
+    (assert fail)
+    (assert warn)
+    (assert (= (funcall fun 1) 1))))
+
+(with-test (:name :typecase-nonfinal-otherwise-warns)
+  (assert-signal
+      (macroexpand-1 '(typecase x (cons 1) (otherwise 2) (t 3)))
+      warning))
+
+(with-test (:name :important-result-through-casts)
+  (assert
+   (nth-value 3
+              (checked-compile `(lambda (x y)
+                                  (the sequence (delete x y))
+                                  10)
+                               :allow-style-warnings t))))
