@@ -446,7 +446,8 @@
                          (pop s))))
       1)))
 
-(with-test (:name :overflow+make-array)
+(with-test (:name :overflow+make-array
+                  :fails-on (not :64-bit))
   (assert
    (= (count 'sb-vm::overflow+t
              (ir2-vops '(lambda (y)
@@ -470,3 +471,79 @@
                                   (optimize speed))
                          (+ a b)))
                      :key (lambda (x) (combination-fun-source-name x nil))))))
+
+(with-test (:name :consecutive-casts)
+  (assert (= (count 'sb-c::%type-check-error/c
+                    (ir-calls
+                     `(lambda (x)
+                        (the fixnum (the integer x))))
+                    :key (lambda (x) (combination-fun-source-name x nil)))
+             1)))
+
+(with-test (:name :sign-extend)
+  (assert (= (count 'sb-c::mask-signed-field
+                    (ir-calls
+                     `(lambda (a)
+                        (declare ((unsigned-byte 32) a))
+                        (logior a (- (mask-field (byte 1 31) a)))))
+                    :key (lambda (x) (combination-fun-source-name x nil)))
+             1))
+  (assert (= (count 'sb-c::mask-signed-field
+                    (ir-calls
+                     `(lambda (a)
+                        (declare ((unsigned-byte 32) a))
+                        (logior (- (mask-field (byte 1 31) a)) a)))
+                    :key (lambda (x) (combination-fun-source-name x nil)))
+             1))
+  #+64-bit
+  (assert (= (count 'sb-c::mask-signed-field
+                    (ir-calls
+                     `(lambda (a)
+                        (declare ((unsigned-byte 64) a))
+                        (logior a (- (mask-field (byte 1 63) a)))))
+                    :key (lambda (x) (combination-fun-source-name x nil)))
+             1)))
+
+
+(with-test (:name :optional-type-checks)
+  (assert (= (count 'sb-c::%type-check-error/c
+                    (ir-calls
+                     `(lambda (&optional x y)
+                        (declare (list x))
+                        (values x y)))
+                    :key (lambda (x) (combination-fun-source-name x nil)))
+             1)))
+
+
+(with-test (:name :flush-multiple-callables)
+  (assert (not (ir-full-calls
+                `(lambda (a b c)
+                   (declare (vector b))
+                   (find a b :test (if c #'eq #'eql))
+                   10)))))
+
+(with-test (:name :stack-allocate-make-array-reverse)
+  (let ((vops (ir2-vops `(lambda (l)
+                           (let ((j (make-array 5 :initial-contents (nreverse l))))
+                             (declare (dynamic-extent j))
+                             (opaque-identity j)
+                             10)))))
+    (assert (= (count 'sb-vm::allocate-vector-on-stack vops) 1))
+    (assert (= (count 'sb-vm::allocate-vector-on-heap vops) 0))))
+
+
+(with-test (:name :no-type-check-tail-call)
+  (destructuring-bind (combination)
+      (ir-full-calls `(lambda (x)
+                        (truly-the fixnum (funcall x))))
+    (assert (node-tail-p combination))))
+
+(with-test (:name :evenp+arithmetic)
+  (assert (not (ir-full-calls `(lambda (x)
+                                 (evenp (+ x 3))))))
+  (assert (not (ir-full-calls `(lambda (x)
+                                 (logbitp 0 (+ x 3)))))))
+
+(with-test (:name :modarith-unknown-types)
+  (assert (not (ir-full-calls `(lambda (x)
+                                 (logand (+ x 10) 20))))))
