@@ -1,4 +1,4 @@
-(unless (gethash 'sb-c:jump-table sb-c::*backend-parsed-vops*)
+(unless (gethash 'sb-c:jump-table sb-c::*backend-template-names*)
   (invoke-restart 'run-tests::skip-file))
 
 (with-test (:name :symbol-case-as-jump-table)
@@ -113,7 +113,7 @@
     (a 'is-a)
     (b 'is-b)
     (c 'is-c)
-    ((or d e) 'is-d-or-e)
+    ((or d e) (print 'is-d-or-e))
     (f 'is-f))))
   (defun typecase-jump-table (x) (guts))
   (defun typecase-no-jump-table (x)
@@ -139,7 +139,7 @@
 
 (with-test (:name :array-subtype-dispatch-table)
   (assert (> (sb-kernel:code-jump-table-words
-              (sb-kernel:fun-code-header #'sb-kernel:vector-subseq*))
+              (sb-kernel:fun-code-header #'sb-kernel:vector-subseq))
              20)))
 
 (with-test (:name :cleanups)
@@ -182,6 +182,7 @@
            ((-28 -41 -38 -99 -110 -81) 0)
            (t 1)))
     (() 1))
+  #+sb-unicode
   (checked-compile
    `(lambda (p1)
       (declare (type character p1))
