@@ -189,6 +189,27 @@
                             (namestring directory))))
     (unless (eq (pathname-host filename) sb-impl::*physical-host*)
       (return-from check-manifest))
+    ;; This special case fixes a new glitch that may occur in debug.impure.lisp.
+    ;; The call sequence in question is:
+    ;;  0xb800b72fbd [RUN-TESTS::CHECK-MANIFEST]
+    ;;  0xb800b77a8b [(LAMBDA (RUN-TESTS::F RUN-TESTS::FILENAME &REST RUN-TESTS::ARGS &KEY :DIRECTION &ALLOW-OTHER-KEYS) :IN RUN-TESTS::PURE-RUNNER)]
+    ;;  0xb800754aba [SB-DI::GET-FILE-TOPLEVEL-FORM]
+    ;;  0xb80075475e [SB-DI::GET-TOPLEVEL-FORM]
+    ;;  0xb800799dca [SB-DEBUG::CODE-LOCATION-SOURCE-FORM]
+    ;;  0xb80079a4cb [SB-DEBUG::LIST-LOCATIONS-DEBUG-COMMAND]
+    ;;  0xb800795597 [SB-DEBUG::DEBUG-LOOP-FUN]
+    ;; It's trying to find something about the file being loaded, but the pathname
+    ;; given to OPEN is STRING/= to *LOAD-PATHNAME*. We should allow it, obviously.
+    ;; Unfortunately if the input-manifest.lisp-expr file itself is missing, then the
+    ;; informational message "Assumed valid input file" is written to *ERROR-OUTPUT*.
+    ;; And that extra output corrupts the running test because the very thing
+    ;; the test asserts on is the contents of *ERROR-OUTPUT*.
+    ;; [why, you may wonder, can the input manifest file be missing? Because the tool
+    ;; which does the sandboxing doesn't send the input manifest itself to the test
+    ;; execution, it only sends the files dictated by the manifest]
+    (let ((lp *load-pathname*))
+      (when (and lp (stem= lp filename)) ; silently permit
+        (return-from check-manifest)))
     (let ((string (namestring filename)))
       (when (or (find #\* (stem-of filename)) ; wild
                 (starts-with-p string "/dev/") ; dev/null and dev/random
@@ -226,6 +247,7 @@
       ,(maybe "SB-IMPL" "*RUN-GC-HOOKS*")
       ,(maybe "SB-VM" "*FNAME-MAP-AVAILABLE-ELTS*")
       ,(maybe "SB-VM" "*FNAME-MAP-OBSERVED-GC-EPOCH*")
+      ,(maybe "SB-UNIX" "*SIGHANDLER-THREAD*")
       sb-impl::**finalizer-store**
       sb-impl::*finalizer-rehashlist*
       sb-impl::*finalizers-triggered*
@@ -265,7 +287,8 @@
       sb-di::*compiled-debug-funs*
       #+win32 sb-impl::*waitable-timer-handle*
       #+win32 sb-impl::*timer-thread*
-      sb-unicode::*name->char-buffers*)))
+      sb-unicode::*name->char-buffers*
+      sb-impl::*finalizer-thread*)))
 
 (defun collect-symbol-values ()
   (let (result)
@@ -392,9 +415,7 @@
     (dolist (package delete)
       (unuse-package (package-use-list package) package))
     ;; Then all deletions
-    (mapc 'delete-package delete)
-    (when delete
-      (format t "::: NOTE: Deleted ~D package~:P~%" (length delete))))
+    (mapc 'delete-package delete))
   ;; Remove PRINT-OBJECT methods specialized on uninterned symbols
   (let ((gf #'print-object))
     (dolist (method (sb-mop:generic-function-methods gf))
@@ -636,23 +657,4 @@
   (filter-test-files "*.impure-cload.lisp"))
 
 (defun sh-files ()
-  (let ((result (filter-test-files "*.test.sh")))
-    #+unix result
-    ;; Rather than hack up the shell scripts which don't pass on #-unix
-    ;; (which would require at least a few lines of shell script and lisp
-    ;; to invoke SBCL and exit with some other code), just confine the kludge
-    ;; to this file.
-    #-unix
-    (if *explicit-test-files*
-        result
-      (remove-if
-       (lambda (x)
-         (member (pathname-name x)
-                 '("filesys.test" ; too many assertions about symlinks to care about just yet
-                   ;; foreign-test-noop-dlclose-test.c:1:10: fatal error: dlfcn.h: No such file or directory
-                   "foreign.test"
-                   ;; No built SBCL here (.../tests/run-sbcl-test-5863): run 'sh make.sh' first!
-                   "run-sbcl.test"
-                   "side-effectful-pathnames.test") ; no idea
-                 :test 'string=))
-        result))))
+  (filter-test-files "*.test.sh"))
