@@ -855,11 +855,11 @@
   (let ((s (make-mock-fd-stream '("zabc" "d" "efgh" "foo*bar" "hi")))
         (string (make-string 100)))
     (let ((endpos
-           (sb-impl::ansi-stream-read-string-from-frc-buffer string s 10 nil)))
+           (read-sequence string s :start 10)))
       (assert (and (= endpos 28)
                    (string= (subseq string 10 endpos) "zabcdefghfoo*barhi"))))
     (let ((endpos
-           (sb-impl::ansi-stream-read-string-from-frc-buffer string s 0 nil)))
+           (read-sequence string s)))
       (assert (= endpos 0)))))
 
 (with-test (:name :named-pipe-wait-eof)
@@ -885,3 +885,60 @@
   (assert (=  (with-open-file (s "/dev/zero")
                 (read-sequence (make-string 4096) s :start 9))
               4096)))
+
+(with-test (:name :io-buffer-sync-basic)
+  (let ((file (scratch-file-name)))
+    (unwind-protect
+         (with-open-file (stream file
+                                 :direction :io
+                                 :if-exists :overwrite
+                                 :if-does-not-exist :create)
+           (write-string "abc" stream)
+           (file-position stream 0)
+           (write-char #\x stream)
+           ;; This should read the character after the 'x' we just wrote,
+           ;; which should be 'b', not 'a' (the original first character)
+           (assert (char= (read-char stream) #\b)))
+      (ignore-errors (delete-file file)))))
+
+(with-test (:name :io-buffer-sync-multiple-ops)
+  (let ((file (scratch-file-name)))
+    (unwind-protect
+         (with-open-file (stream file
+                                 :direction :io
+                                 :if-exists :overwrite
+                                 :if-does-not-exist :create)
+           (write-string "hello world" stream)
+           (file-position stream 6)  ; position after "hello "
+           (write-string "SBCL" stream)
+           (file-position stream 9)  ; position at last character of "SBCL"
+           ;; Should read the 'L' from "SBCL", not 'r' from original "world"
+           (assert (char= (read-char stream) #\L)))
+      (ignore-errors (delete-file file)))))
+
+#+unix
+(with-test (:name :unbuffered-nonblocking-write-sequence)
+  (flet ((make-non-blocking-pipe ()
+           (multiple-value-bind (i o) (sb-posix:pipe)
+             (sb-posix:fcntl i sb-posix:f-setfl sb-posix:o-nonblock)
+             (sb-posix:fcntl o sb-posix:f-setfl sb-posix:o-nonblock)
+             (make-two-way-stream (sb-sys:make-fd-stream i :input t :buffering :none :element-type '(unsigned-byte 8))
+                                  (sb-sys:make-fd-stream o :output t :buffering :none :element-type '(unsigned-byte 8))))))
+    (let ((pipe (make-non-blocking-pipe))
+          (buf (make-array 1 :element-type '(unsigned-byte 8))))
+      (unwind-protect
+           (progn
+             ;; Fill the pipe buffer
+             (loop while (sb-unix:unix-write (sb-sys:fd-stream-fd (two-way-stream-output-stream pipe)) buf 0 1))
+             (write-sequence (coerce #(1 2 3 4) '(vector (unsigned-byte 8))) pipe)
+             (let ((in (make-array 65536 :element-type '(unsigned-byte 8)
+                                         :initial-element 255)))
+               (read-sequence in pipe)
+               (assert (every #'zerop in)))
+             (write-sequence (make-array 4 :element-type '(unsigned-byte 8) :initial-element 0) pipe)
+             (let ((in (make-array 4 :element-type '(unsigned-byte 8)
+                                     :initial-element 255)))
+               (read-sequence in pipe)
+               (assert (equalp in #(1 2 3 4)))))
+        (close (two-way-stream-output-stream pipe))
+        (close (two-way-stream-input-stream pipe))))))
