@@ -458,7 +458,8 @@
                      (= safety 3))))
     ((eql :safe)
      (list :filter (lambda (&key speed safety &allow-other-keys)
-                     (and (> safety 0) (>= safety speed)))))
+                     (and (> safety 0) (>= safety speed)))
+           :compilation-speed 1 :space 1))
     ((eql :quick)
      '(:compilation-speed 1 :space 1))
     ((eql :quick/incomplete)
